/* Parham Data Products                                                      */
/* Expert Rules Engine                                                       */
/* main/expert/model/rulesEngine/InferenceEngine                             */
/* Copyright 2020 Wayne Parham                                               */

package com.parhamdata.expert.model.rulesEngine;

import com.parhamdata.expert.services.databaseService.Rule;
import com.parhamdata.expert.model.rulesEngine.langParser.RuleParser;
import com.parhamdata.expert.model.RulesetDomain;
import org.springframework.stereotype.Service;
import org.springframework.beans.factory.annotation.Autowired;
import java.util.stream.Collectors;
import java.util.Comparator;
import java.util.ArrayList;
import java.util.Optional;
import java.util.List;

@Service
public abstract class InferenceEngine<Inputs, Results> {

    @Autowired
    private RuleParser<Inputs, Results> ruleParser;

    public Object run(List<Rule> listOfRules, Inputs inputData, boolean multipleResultsAllowed) {
        Results       outputResult     = null;
        List<Rule>    matchingRules    = null;
        Rule          resolvedRule     = null;

        if (null == listOfRules || listOfRules.isEmpty()) {
            return null;
        }

        initializeOutputResult();
        matchingRules = match(listOfRules, inputData);

        if (!multipleResultsAllowed) {
            resolvedRule = resolve(matchingRules);

            if (null == resolvedRule) {
                return null;
            }
            outputResult = executeRule(resolvedRule, inputData);
            return outputResult;

        } else {
            ArrayList<Results> outputResults = new ArrayList<Results>();
            for(Rule rule : matchingRules) {
                outputResult = executeRule(rule, inputData);
                outputResults.add(outputResult);
            }
            return outputResults;
        }
    }

    protected List<Rule> match(List<Rule> listOfRules, Inputs inputData) {
        return listOfRules.stream()
                .filter(
                        rule -> {
                            String condition = rule.getCondition();
                            return ruleParser.parseCondition(condition, inputData);
                        }
                )
                .collect(Collectors.toList());
    }

    protected Rule resolve(List<Rule> conflictSet) {
        Optional<Rule> rule = conflictSet.stream()
                .sorted(Comparator.comparingInt(Rule::getPriority))
                .findFirst();
        if (rule.isPresent()) {
            return rule.get();
        }
        return null;
    }

    protected Results executeRule(Rule rule, Inputs inputData) {
        Results outputResult = initializeOutputResult();
        return ruleParser.parseAction(rule.getAction(), inputData, outputResult);
    }

    protected abstract Results initializeOutputResult();
    protected abstract RulesetDomain getDomain();
}